/* binclock - see szInfo below for copyright and licence details */

const char szInfo[] = "\
Binary Clock\n\
Version 2.0\n\
\n\
A binary clock and stopwatch that runs in the Windows system tray.\n\
\n\
Copyright (c) 2004-2005 Andrew Wedgbury <wedge@sconemad.com>\n\
\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2, or (at your option)\n\
any later version.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program (see the file COPYING); if not, write to the\n\
Free Software Foundation, Inc.,\n\
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA\n\
";

#include <windows.h>
#include <stdio.h>
#include <string>

// Const strings
const char szAppName[] = "Binary Clock";
const char szWindowClass[] = "BinClock class";
const char szRegKey[] = "Software\\SconeMAD\\binclock";

// Global pointer to the view root object, everything needs to access this
class BCViewRoot;
BCViewRoot* root;

// This is updated with the current time
SYSTEMTIME current_time;

// Maximum number of views
const int MAX_VIEWS = 10;

// Command IDs
#define IDC_ABOUT 101
#define IDC_EXIT 102

#define IDC_SET_ON_COL 201
#define IDC_SET_OFF_COL 202

#define IDC_RUN 301
#define IDC_RESET 302

#define IDC_MODE 400

#define IDC_SET_LIMIT 1000

// Prototypes
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Base view class
//
class BCView {
public:
  BCView();
  virtual ~BCView();

  virtual bool load(HKEY key);
  virtual bool save(HKEY key);
  // Load/save settings from specified registry key

  virtual std::string name() =0;
  // Get the name of this view

  virtual void menu(HMENU hmenu);
  // Construct context menu for this view

  virtual void command(int id);
  // Process command for this view

  virtual void count(DWORD delta);
  // Perform any required counting for this view

  virtual int value();
  // Get the value to display for this view (should be 0-9999)

  COLORREF m_col_on;
  COLORREF m_col_off;
  // Colours to use for this view's display
};

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Time view class - displays the current time (HH:MM)
//
class BCViewTime : public BCView {
public:
  virtual std::string name();
  virtual void menu(HMENU hmenu);
  virtual void command(int id);
  virtual int value();
};

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Date view class - displays the current date (mm:dd)
//
class BCViewDate : public BCView {
public:
  virtual std::string name();
  virtual void menu(HMENU hmenu);
  virtual void command(int id);
  virtual int value();
};

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Stopwatch view class - displays and controls stopwatch timer (MM:SS)
//
class BCViewStopwatch : public BCView {
public:
  BCViewStopwatch();

  virtual std::string name();
  virtual void menu(HMENU hmenu);
  virtual void command(int id);
  virtual void count(DWORD delta);
  virtual int value();

private:
  long int m_value;
  bool m_running;
};

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Countdown view class - displays and controls countdown timer (MM:SS)
//
class BCViewCountdown : public BCView {
public:
  BCViewCountdown();

  virtual std::string name();
  virtual void menu(HMENU hmenu);
  virtual void command(int id);
  virtual void count(DWORD delta);
  virtual int value();

private:
  long int m_value;
  int m_start;
  bool m_running;
};

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Fancy view class - displays a fancy pattern
//
class BCViewFancy : public BCView {
public:
  BCViewFancy();

  virtual std::string name();
  virtual void count(DWORD delta);
  virtual int value();

private:
  long int m_value;
};

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// View root - controls the view objects.
//
class BCViewRoot {
public:
  BCViewRoot(HINSTANCE hinst);
  ~BCViewRoot();

  bool init();
  bool close();
  std::string name();
  void left_click();
  void left_dblclick();
  void right_click();
  void count(DWORD delta);
  int value();
  void command(int id);
  void add_view(BCView* v);
  BCView* current_view();
  void alert();
  void update_icon();

  HINSTANCE m_hinstance;
  HWND m_hwnd;
  HMENU m_hmenu;
  NOTIFYICONDATA m_nid;
  HBITMAP m_hbm_icon;
  HBITMAP m_hbm_mask;
  HDC m_hdc_icon;
  HDC m_hdc_mask;
  HBRUSH m_hbr_on;
  HBRUSH m_hbr_off;
  COLORREF m_col_custom[16];
  int m_icon_x;
  int m_icon_y;
  int m_current_view;
  int m_num_views;
  BCView* m_views[MAX_VIEWS];
  bool m_alert;
}; 

//=============================================================================
BCView::BCView()
  : m_col_on(RGB(255,255,255)),
    m_col_off(RGB(0,0,0))
{
  
}
  
//=============================================================================
BCView::~BCView() 
{
  
}

//=============================================================================
bool BCView::load(HKEY key) 
{
  for (int c=0; c<2; ++c) {
    COLORREF& col = (c ? m_col_on : m_col_off);
    DWORD type;
    char buffer[10];
    unsigned long size = sizeof(buffer);
    char ent[32];
    sprintf(ent,"%s_col_%s",name().c_str(),c?"on":"off");
    if (ERROR_SUCCESS ==
	RegQueryValueEx(key,ent,0,&type,(BYTE*)buffer,&size)
	&& size == 7) {
      BYTE cvec[3] = {0,0,0};
      for (int v=0; v<3; ++v) {
	char b[3] = {buffer[2*v],buffer[(2*v)+1],'\0'};
	int i;
	if (sscanf(b,"%x",&i)) {
	  cvec[v] = (BYTE)i;
	} else {
	  break;
	}
      }
      col = RGB(cvec[0],cvec[1],cvec[2]);
    }
  }
  return true;
}

//=============================================================================
bool BCView::save(HKEY key) 
{
  for (int c=0; c<2; ++c) {
    COLORREF& col = (c ? m_col_on : m_col_off);
    char buffer[10];
    char ent[32];
    sprintf(ent,"%s_col_%s",name().c_str(),c?"on":"off");
    sprintf(buffer,"%02x%02x%02x",
	    GetRValue(col),GetGValue(col),GetBValue(col));
    RegSetValueEx(key,ent,0,REG_SZ,(BYTE*)buffer,lstrlen(buffer)+1);
  }
  return true;
}

//=============================================================================
void BCView::menu(HMENU hmenu)
{
  AppendMenu(hmenu,MF_STRING,IDC_SET_ON_COL,"Set &On colour");
  AppendMenu(hmenu,MF_STRING,IDC_SET_OFF_COL,"Set O&ff colour");
  AppendMenu(hmenu,MF_SEPARATOR,0,0);
}

//=============================================================================
void BCView::command(int id)
{
  if (id == IDC_SET_ON_COL || 
      id == IDC_SET_OFF_COL) {
    // Set on/off colours
    COLORREF& col = (id == IDC_SET_ON_COL ? m_col_on : m_col_off);
    CHOOSECOLOR cc;
    int size = sizeof(CHOOSECOLOR);
    FillMemory(&cc,size,0x00);
    cc.lStructSize = size;
    cc.hwndOwner = root->m_hwnd;
    cc.rgbResult = col;
    cc.lpCustColors = root->m_col_custom;
    cc.Flags = CC_RGBINIT;
    if (ChooseColor(&cc)) {
      col = cc.rgbResult;
    }
  }
}

//=============================================================================
void BCView::count(DWORD delta)
{
}

//=============================================================================
int BCView::value()
{
  return 0;
}

//=============================================================================
std::string BCViewTime::name()
{ 
  return "Time"; 
}

//=============================================================================
void BCViewTime::menu(HMENU hmenu)
{
  BCView::menu(hmenu);
  AppendMenu(hmenu,MF_STRING,IDC_RESET,"Set date/time");
}

//=============================================================================
void BCViewTime::command(int id)
{
  switch (id) {
    case IDC_RESET:
      WinExec("control.exe date/time",SW_NORMAL);
      break;
    default:
      BCView::command(id);
      break;
  }
}

//=============================================================================
int BCViewTime::value() 
{
  return (100*current_time.wHour) + current_time.wMinute;
}

//=============================================================================
std::string BCViewDate::name()
{ 
  return "Date"; 
}

//=============================================================================
void BCViewDate::menu(HMENU hmenu)
{
  BCView::menu(hmenu);
  AppendMenu(hmenu,MF_STRING,IDC_RESET,"Set date/time");
}

//=============================================================================
void BCViewDate::command(int id)
{
  switch (id) {
    case IDC_RESET:
      WinExec("control.exe date/time",SW_NORMAL);
      break;
    default:
      BCView::command(id);
      break;
  }
}

//=============================================================================
int BCViewDate::value()
{
  return (100*current_time.wMonth) + current_time.wDay;
}

//=============================================================================
BCViewStopwatch::BCViewStopwatch()
  : m_value(0),
    m_running(0)
{
}

//=============================================================================
std::string BCViewStopwatch::name() 
{ 
  return "Stopwatch"; 
}

//=============================================================================
void BCViewStopwatch::menu(HMENU hmenu)
{
  BCView::menu(hmenu);
  AppendMenu(hmenu,MF_STRING,IDC_RESET,"Reset");
  if (m_running) {
    AppendMenu(hmenu,MF_STRING,IDC_RUN,"Stop");
  } else {
    AppendMenu(hmenu,MF_STRING,IDC_RUN,"Start");
  }
}

//=============================================================================
void BCViewStopwatch::command(int id)
{
  switch (id) {
    case IDC_RESET:
      m_running = false;
      m_value = 0;
      break;
    case IDC_RUN:
      m_running = !m_running;
      break;
    default:
      BCView::command(id);
      break;
  }
}

//=============================================================================
void BCViewStopwatch::count(DWORD delta)
{
  if (m_running) {
    m_value += delta;
  }
}
  
//=============================================================================
int BCViewStopwatch::value() 
{
  int m = m_value / 60000;
  int s = (m_value/1000) % 60;
  return (100*m) + s;
}

//=============================================================================
BCViewCountdown::BCViewCountdown()
  : m_value(0),
    m_start(5*60),
    m_running(0)
{
}
  
//=============================================================================
std::string BCViewCountdown::name() 
{ 
  return "Countdown"; 
}

//=============================================================================
void BCViewCountdown::menu(HMENU hmenu)
{
  BCView::menu(hmenu);

  HMENU hsubmenu = CreatePopupMenu();

  AppendMenu(hsubmenu,MF_STRING | (30==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + 10,"10 seconds");
  AppendMenu(hsubmenu,MF_STRING | (20==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + 20,"20 seconds");
  AppendMenu(hsubmenu,MF_STRING | (30==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + 30,"30 seconds");
  AppendMenu(hsubmenu,MF_STRING | (40==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + 40,"40 seconds");
  AppendMenu(hsubmenu,MF_STRING | (50==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + 50,"50 seconds");

  AppendMenu(hsubmenu,MF_STRING | (60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (1*60),"1 minute");
  AppendMenu(hsubmenu,MF_STRING | (2*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (2*60),"2 minutes");
  AppendMenu(hsubmenu,MF_STRING | (3*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (3*60),"3 minutes");
  AppendMenu(hsubmenu,MF_STRING | (4*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (4*60),"4 minutes");
  AppendMenu(hsubmenu,MF_STRING | (5*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (5*60),"5 minutes");

  AppendMenu(hsubmenu,MF_STRING | (10*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (10*60),"10 minutes");
  AppendMenu(hsubmenu,MF_STRING | (20*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (20*60),"20 minutes");
  AppendMenu(hsubmenu,MF_STRING | (30*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (30*60),"30 minutes");
  AppendMenu(hsubmenu,MF_STRING | (40*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (40*60),"40 minutes");
  AppendMenu(hsubmenu,MF_STRING | (50*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (50*60),"50 minutes");

  AppendMenu(hsubmenu,MF_STRING | (60*60==m_start ? MF_CHECKED : 0),
	     IDC_SET_LIMIT + (60*60),"1 hour");

  AppendMenu(hmenu,MF_POPUP,(UINT_PTR)hsubmenu,"Timeout");

  AppendMenu(hmenu,MF_STRING,IDC_RESET,"Reset");
  if (m_running) {
    AppendMenu(hmenu,MF_STRING,IDC_RUN,"Stop");
  } else {
    AppendMenu(hmenu,MF_STRING,IDC_RUN,"Start");
  }
}

//=============================================================================
void BCViewCountdown::command(int id)
{
  if (id > IDC_SET_LIMIT) {
    m_start = id - IDC_SET_LIMIT;

  } else {

    switch (id) {
      case IDC_RESET:
	m_running = false;
	m_value = 0;
	break;
      case IDC_RUN:
	m_running = !m_running;
	if (m_value == 0) {
	  m_value = m_start * 1000;
	}
	break;
      default:
	BCView::command(id);
	break;
    }
  }
}

//=============================================================================
void BCViewCountdown::count(DWORD delta)
{
  if (m_running) {
    m_value -= delta;
    if (m_value <= 0) {
      m_running = false;
      m_value = 0;
      root->alert();
    }
  }
}
  
//=============================================================================
int BCViewCountdown::value() 
{
  int m = m_value / 60000;
  int s = (m_value/1000) % 60;
  return (100*m) + s;
}

//=============================================================================
BCViewFancy::BCViewFancy()
  : m_value(0)
{
}

//=============================================================================
std::string BCViewFancy::name() 
{ 
  return "Random"; 
}


//=============================================================================
void BCViewFancy::count(DWORD delta)
{
  ++m_value;
}


//=============================================================================
int BCViewFancy::value() 
{
  int r=0;
  r += MulDiv(rand(),8,32768);
  r += 10 * MulDiv(rand(),8,32768);
  r += 100 * MulDiv(rand(),8,32768);
  r += 1000 * MulDiv(rand(),8,32768);
  return r;
}

//=============================================================================
BCViewRoot::BCViewRoot(HINSTANCE hinst) 
  : m_hinstance(hinst),
    m_current_view(0),
    m_num_views(0),
    m_alert(false)
{
  // Initialise views array to NULLs
  for (int i=0; i<MAX_VIEWS; ++i) {
    m_views[i] = 0;
  }
}
  
//=============================================================================
BCViewRoot::~BCViewRoot() 
{
  // Delete all views
  for (int i=0; i<m_num_views; ++i) {
    delete m_views[i];
  }
}

//=============================================================================
bool BCViewRoot::init() 
{
  WNDCLASSEX wc;
  int size = sizeof(WNDCLASSEX);
  FillMemory(&wc,size,0x00);
  wc.cbSize = size;
  wc.lpfnWndProc  = (WNDPROC)WndProc;
  wc.hInstance    = m_hinstance;
  wc.lpszClassName= szWindowClass;
  RegisterClassEx(&wc);
  m_hwnd = CreateWindow(szWindowClass,szAppName,WS_OVERLAPPEDWINDOW,
			0,0,0,0,NULL,NULL,m_hinstance,NULL);
  if (!m_hwnd) {
    return false;
  }
  
  m_icon_x = GetSystemMetrics(SM_CXSMICON);
  m_icon_y = GetSystemMetrics(SM_CYSMICON);
  
  // Create the bitmaps used to draw the icon and mask
  HDC hdc_screen = GetDC(0);
  m_hdc_icon = CreateCompatibleDC(hdc_screen);
  m_hbm_icon = CreateCompatibleBitmap(hdc_screen,m_icon_x,m_icon_y);
  m_hdc_mask = CreateCompatibleDC(hdc_screen);
  m_hbm_mask = CreateBitmap(m_icon_x,m_icon_y,1,1,0);
  
  // Load registry settings for each view    
  HKEY key=0;
  RegOpenKeyEx(HKEY_CURRENT_USER,szRegKey,0,KEY_READ,&key);
  for (int i=0; i<m_num_views; ++i) {
    m_views[i]->load(key);
  }
  RegCloseKey(key);
  
  // Create the systray icon
  m_nid.cbSize = sizeof(NOTIFYICONDATA);
  m_nid.uFlags = NIF_MESSAGE;
  m_nid.hWnd = m_hwnd;
  m_nid.uCallbackMessage = WM_USER;
  Shell_NotifyIcon(NIM_ADD,&m_nid);
  
  return true;
}

//=============================================================================
bool BCViewRoot::close() 
{
  // Delete bitmaps and DCs
  DeleteDC(m_hdc_icon);
  DeleteObject(m_hbm_icon);
  DeleteDC(m_hdc_mask);
  DeleteObject(m_hbm_mask);
  
  Shell_NotifyIcon(NIM_DELETE,&m_nid);
  
  HKEY key;
  DWORD disp;
  if (ERROR_SUCCESS !=
      RegCreateKeyEx(HKEY_CURRENT_USER,szRegKey,0,0,
		     REG_OPTION_NON_VOLATILE,KEY_WRITE,0,&key,&disp)) {
    return false;
  }
  for (int i=0; i<m_num_views; ++i) {
    m_views[i]->save(key);
  }
  RegCloseKey(key);
  return true;
}

//=============================================================================
std::string BCViewRoot::name()
{
  return current_view()->name();
}

//=============================================================================
void BCViewRoot::left_click()
{
  if (m_alert) {
    m_alert = false;
  } else {
    current_view()->command(IDC_RUN);
  }
  update_icon();
}
 
//=============================================================================
void BCViewRoot::left_dblclick()
{
  if (m_alert) {
    m_alert = false;
  } else {
    current_view()->command(IDC_RESET);
  }
  update_icon();
}

//=============================================================================
void BCViewRoot::right_click()
{
  // Popup menu when icon is right-clicked
  POINT pt;
  GetCursorPos(&pt);
  if (m_hmenu) {
    DestroyMenu(m_hmenu);
  }
  m_hmenu = CreatePopupMenu();
  AppendMenu(m_hmenu,MF_STRING,IDC_ABOUT,"&About");
  AppendMenu(m_hmenu,MF_SEPARATOR,0,0);

  // Allow the current view to add menu items
  current_view()->menu(m_hmenu);
  AppendMenu(m_hmenu,MF_SEPARATOR,0,0);

  // Add view selectors
  for (int i=0; i<m_num_views; ++i) {
    char name[32];
    sprintf(name,"%s\t[%02d:%02d]",
	    m_views[i]->name().c_str(),
	    m_views[i]->value() / 100,
	    m_views[i]->value() % 100);
    AppendMenu(m_hmenu,
	       MF_STRING | (i==m_current_view ? MF_CHECKED : 0),
	       IDC_MODE + i,
	       name);
  }
  AppendMenu(m_hmenu,MF_SEPARATOR,0,0);
  AppendMenu(m_hmenu,MF_STRING,IDC_EXIT,"&Exit");

  // Add exit 
  
  SetForegroundWindow(m_hwnd);
  TrackPopupMenu(m_hmenu,0,pt.x,pt.y,0,m_hwnd,0);
  PostMessage(m_hwnd,WM_NULL,0,0);
}

//=============================================================================
void BCViewRoot::count(DWORD delta)
{
  for (int i=0; i<m_num_views; ++i) {
    m_views[i]->count(delta);
  }
  update_icon();
}

//=============================================================================
int BCViewRoot::value()
{
  return current_view()->value();
}

//=============================================================================
void BCViewRoot::command(int id)
{
  if (id >= IDC_MODE && id < IDC_MODE+m_num_views) {
    m_current_view = id-IDC_MODE;
  } else {
    current_view()->command(id);
  }
}

//=============================================================================
void BCViewRoot::add_view(BCView* v)
{
  m_views[m_num_views] = v;
  ++m_num_views;
}

//=============================================================================
BCView* BCViewRoot::current_view()
{
  return m_views[m_current_view];
}

//=============================================================================
void BCViewRoot::alert()
{
  m_alert = true;
  MessageBeep(MB_ICONASTERISK);
}

//=============================================================================
void BCViewRoot::update_icon()
{
  // Split view's value into individual digits
  int  v = value();
  unsigned int a[4];
  a[0] = v / 1000;
  a[1] = (v - a[0]*1000) / 100;
  a[2] = (v - a[0]*1000 - a[1]*100) / 10;
  a[3] = (v - a[0]*1000 - a[1]*100 - a[2]*10); 

  // Get the icon's rectangle
  RECT r;
  r.top = 0; r.bottom = m_icon_y;
  r.left = 0; r.right = m_icon_x;
  
  // Clear icon and mask
  HGDIOBJ save_icon = SelectObject(m_hdc_icon,m_hbm_icon);
  FillRect(m_hdc_icon,&r,(HBRUSH)GetStockObject(WHITE_BRUSH));
  HGDIOBJ save_mask = SelectObject(m_hdc_mask,m_hbm_mask);
  FillRect(m_hdc_mask,&r,(HBRUSH)GetStockObject(WHITE_BRUSH));

  // Get the colours required for drawing.
  // Reverse colours for alternate seconds if alert condition is active.
  bool flash = (m_alert && (current_time.wSecond % 2));
  COLORREF on = flash ? 
    current_view()->m_col_off : current_view()->m_col_on;
  COLORREF off = flash ? 
    current_view()->m_col_on : current_view()->m_col_off;
  
  // Create brushes for painting
  HBRUSH hbr_on = CreateSolidBrush(on);
  HBRUSH hbr_off = CreateSolidBrush(off);
  
  int xp = (m_icon_x/4);
  int yp = (m_icon_y/4);
  
  // Draw the pattern
  for (int x=0; x<4; ++x) {
    for (int y=0; y<4; ++y) {
      r.top = y*yp; r.bottom = r.top+yp-1;
      r.left = x*xp; r.right = r.left+xp-1;
      bool on = 0 != (a[x]&(8>>y));
      FillRect(m_hdc_icon,&r,on ? hbr_on : hbr_off);
      FillRect(m_hdc_mask,&r,(HBRUSH)GetStockObject(BLACK_BRUSH));
    }
  }
  
  SelectObject(m_hdc_icon,save_icon);
  SelectObject(m_hdc_mask,save_mask);
  
  // Delete brushes
  DeleteObject(hbr_on);
  DeleteObject(hbr_off);
    
  // Set the tool tip to the current date
  GetDateFormat(LOCALE_USER_DEFAULT,LOCALE_NOUSEROVERRIDE | DATE_LONGDATE,
		&current_time,NULL,m_nid.szTip,64);

  // Construct a new icon from the bitmap and mask
  ICONINFO iin;
  iin.fIcon = TRUE;
  iin.hbmMask = m_hbm_mask;
  iin.hbmColor = m_hbm_icon;

  // Update the systray icon
  HICON hicon = CreateIconIndirect(&iin);
  m_nid.hIcon = hicon;
  m_nid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
  Shell_NotifyIcon(NIM_MODIFY,&m_nid);

  // Clean up
  DestroyIcon(hicon);
}

//===========================================================================
// Main window class
//
LRESULT CALLBACK WndProc(
  HWND hWnd,
  UINT message,
  WPARAM wParam,
  LPARAM lParam
) 
{
  // Store the previous tick count, so we can work out delta.
  static DWORD prev_ticks;

  switch (message) {

    case WM_CREATE: {
      // Use a 1 second timer to update the icon
      prev_ticks = GetTickCount();
      SetTimer(hWnd,1,250,0);
    } return 0;

    case WM_TIMER: {
      // Update current time variable
      GetLocalTime(&current_time);

      // Work out number of ticks since last timer event and pass on to view.
      DWORD current_ticks = GetTickCount();
      // (handle rare timer rollover condition)
      if (current_ticks < prev_ticks) prev_ticks = 0;
      root->count(current_ticks - prev_ticks);
      prev_ticks = current_ticks;
    } break;

    case WM_USER: {
      // Process clicky-type messages from the icon by passing them onto view.
      switch (lParam) {
        case WM_LBUTTONDOWN:
	  root->left_click();
	  break;
        case WM_RBUTTONDOWN:
	  root->right_click();
	  break;
        case WM_LBUTTONDBLCLK:
	  root->left_dblclick();
	  break;
      }
    } break;

    case WM_COMMAND: {
      // Handle command messages from menu. Some are handled here, the rest
      // are passed onto the view.
      int id = LOWORD(wParam);
      switch (id) {
        case IDC_ABOUT:
          MessageBox(hWnd,szInfo,szAppName,MB_OK | MB_ICONINFORMATION);
	  { 
	    static int about_count=0;
	    if (++about_count == 5) {
	      root->add_view(new BCViewFancy);
	    }
	  }
          break;
        case IDC_EXIT:
          PostMessage(hWnd,WM_DESTROY,0,0);
          break;
        default:
	  root->command(id);
	  break;
      }
    } break;

    case WM_DESTROY: {
      // App is shutting down, stop the timer.
      KillTimer(hWnd,1);
      PostQuitMessage(0);
    } break;

    default:
      return DefWindowProc(hWnd, message, wParam, lParam);
  }
  return 0;
}

//===========================================================================
// ENTRY POINT
//
int APIENTRY WinMain(
  HINSTANCE hInstance,
  HINSTANCE hPrevInstance,
  LPSTR lpCmdLine,
  int nCmdShow
)
{
  // Create root
  root = new BCViewRoot(hInstance);

  // Add required views
  root->add_view(new BCViewTime);
  root->add_view(new BCViewDate);
  root->add_view(new BCViewStopwatch);
  root->add_view(new BCViewCountdown);

  // Init views - this will create the main window (which is hidden) and 
  // systray icon, then read settings from the registry for each view.
  if (root->init() == false) {
    MessageBox(0,"Error occured when starting the application",
	       szAppName,MB_ICONWARNING | MB_OK);
    return 1;
  }

  // Message loop
  MSG msg;
  while (GetMessage(&msg, NULL, 0, 0)) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  
  // Close views and clean up - This will save current settings to the
  // registry.
  if (root->close() == false) {
    MessageBox(0,"Error occured when saving settings",
	       szAppName,MB_ICONWARNING | MB_OK);
  }
  delete root;

  // Return to system
  return msg.wParam;
}
